﻿using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using AviReader;

namespace SlitTears2 {
    public partial class Form1 : Form {

        Bitmap[] videoFrames;
        TimelineForm timelineForm;
        SlitTool slitTool;
        public enum SlitTool {
            Line, Squiggly
        }

        Point lineStart, lineEnd;
        bool drawLine;
        bool drawSquiggly;

        // TODO: Do this with 2DPath
        List<Point> slits;
        List<Point> preSlit;


        public Form1() {
            InitializeComponent();
            this.timelineForm = new TimelineForm();
            this.timelineForm.TimelineIndexChanged += new EventHandler(timelineForm_TimelineIndexChanged);
            this.slitTool = SlitTool.Line;
            this.pictureBox1.Paint += new PaintEventHandler(pictureBox1_Paint);
            this.slits = new List<Point>();
            this.preSlit = new List<Point>();
        }

        /// <summary>
        /// Updates the video frame with the slit tears.
        /// </summary>
        protected void pictureBox1_Paint(object sender, PaintEventArgs e) {
            // Draws the "pre-slit stuff" -- i.e. where the user has indicated s/he wants a slit-tear
            if (this.drawLine) {
                using (Pen p = new Pen(Color.FromArgb(100, Color.Green), 5)) {
                    using (Pen p2 = new Pen(Color.Green)) {
                        e.Graphics.DrawLine(p, this.lineStart, this.lineEnd);
                        e.Graphics.DrawLine(p2, this.lineStart, this.lineEnd);
                    }
                }
            }
            else if (this.drawSquiggly) {
                foreach (Point p in this.preSlit) {
                    e.Graphics.DrawRectangle(Pens.Green, p.X, p.Y, 1, 1);
                }                
            }

            // Draw the real slits
            Color slitColor = Color.FromArgb(150, Color.Red);
            using (Pen pen = new Pen(slitColor)) {
                foreach (Point p in this.slits) {
                    e.Graphics.DrawRectangle(pen, p.X, p.Y, 1, 1);
                }
            }

        }

        /// <summary>
        /// Starts the openFileDialog, and loads in the frames from the video file.
        /// </summary>
        private void button1_Click(object sender, EventArgs e) {
            int framesToGrab = 0;
            int leadingSkipFrames = (int) this.numericUpDown3.Value;
            int framesToSkip = (int) this.numericUpDown4.Value;
            int totalVideoFrames = 0;

            if (DialogResult.OK != this.openFileDialog1.ShowDialog(this)) {
                return;
            }

            try {
                AviReader.AviReader aviReader = new AviReader.AviReader();
                aviReader.Open(this.openFileDialog1.FileName);

                // How many video frames are left after we skip the leading frames
                totalVideoFrames = aviReader.Count - leadingSkipFrames;

                if (totalVideoFrames < 0) throw new Exception("Not enough video frames in file...");

                // Make sure we're not trying to grab more frames than are in the actual file
                // ... given the skipping pattern
                framesToGrab = Math.Min(totalVideoFrames / (framesToSkip + 1), (int)this.numericUpDown1.Value);

                int startAt = leadingSkipFrames;
                int endAt = leadingSkipFrames + framesToGrab * (framesToSkip + 1);
                int increment = framesToSkip + 1;

                this.videoFrames = new Bitmap[framesToGrab];
                this.progressBar1.Maximum = framesToGrab;
                this.progressBar1.Value = 0;

                // TODO: Throw this to a background worker?
                //for (int i = 0; i < this.videoFrames.Length; i++) {
                //    this.Text = "Loading frame: " + i + "/" + this.videoFrames.Length;
                //    Bitmap bmp = aviReader.ReadFrame(i);
                //    this.videoFrames[i] = bmp;
                //    this.progressBar1.Increment(1);
                //}
                for (int i = startAt, j = 0; i < endAt; i += increment, j++) {
                    this.Text = "Loading frame: " + j + "/" + this.videoFrames.Length;
                    Bitmap bmp = aviReader.ReadFrame(i);
                    this.videoFrames[j] = bmp;
                    this.progressBar1.Increment(1);
                }
                this.Text = "SlitTears: " + Path.GetFileName(this.openFileDialog1.FileName) + "; " + this.videoFrames.Length + " frames - " +
                    ((startAt > 0) ? "skipped: " + startAt + " frames;" : "") + ((increment > 1) ? " loaded 1/" + increment + " frames" : "");

                // Set up the TimeLine window
                this.timelineForm.TimelineSize = new Size(this.videoFrames.Length, (int)this.numericUpDown2.Value);
                this.timelineForm.TimelineIndex = 0;
                this.timelineForm.Show();
                this.timelineForm.Location = this.Location + new Size(0, this.Height);

                this.slits.Clear();
            }
            catch {
                MessageBox.Show("Can't load the video file... :(");
            }
        }

        /// <summary>
        /// Callback from the timeline window: this updates the picturebox to show the correct
        /// frame in the video.
        /// </summary>
        protected void timelineForm_TimelineIndexChanged(object sender, EventArgs e) {
            if (this.videoFrames == null) return;
            this.pictureBox1.Image = this.videoFrames[this.timelineForm.TimelineIndex];
        }

        /// <summary>
        /// Switches which slitTool the user is using
        /// </summary>
        private void changeSlitTool(object sender, EventArgs e) {
            if (sender == radioButton1) {
                this.slitTool = SlitTool.Line;
            }
            else if (sender == radioButton2) {
                this.slitTool = SlitTool.Squiggly;
            }
        }

        /// <summary>
        /// Start collecting information about where the user wants a slit tear
        /// </summary>
        private void pictureBox1_MouseDown(object sender, MouseEventArgs e) {
            if (e.Button == MouseButtons.Right) {
                this.resetTimeline();
                return;
            }

            if (this.slitTool == SlitTool.Line) {
                
                // For the line, just collect start and end points
                this.lineStart = e.Location;
                this.lineEnd = e.Location;
                this.drawLine = true;
                this.pictureBox1.Refresh();
            }
            else if (this.slitTool == SlitTool.Squiggly) {

                // For the squiggly, collect start and end points for each part, and connect
                // each part using bresenham
                this.lineStart = e.Location;
                this.preSlit.Clear();
                this.preSlit.Add(e.Location);
                this.drawSquiggly = true;
                this.pictureBox1.Refresh();
            }
        }

        /// <summary>
        /// Update the pre-slit information, and refresh the picturebox
        /// </summary>
        private void pictureBox1_MouseMove(object sender, MouseEventArgs e) {
            if (e.Button != MouseButtons.Left) return;

            if (this.slitTool == SlitTool.Line) {
                this.lineEnd = e.Location;
                this.pictureBox1.Refresh();
            }
            else if (this.slitTool == SlitTool.Squiggly) {
                
                // Store the pixel points for this segment in preSlit
                this.bresenhamDrawLine(this.lineStart, e.Location, this.preSlit);
                this.lineStart = e.Location;
                this.pictureBox1.Refresh();
            }
        }

        /// <summary>
        /// Using pre-slit information, actually take the information from the video frames 
        /// and put into timeline.  In both these cases, we call updateTimeline with just the
        /// number of points created by this mouse-stroke.
        /// </summary>
        private void pictureBox1_MouseUp(object sender, MouseEventArgs e) {
            if (e.Button != MouseButtons.Left) return;

            if (this.slitTool == SlitTool.Line) {
                this.lineEnd = e.Location;
                this.drawLine = false;

                int pointCount = this.slits.Count;
                this.bresenhamDrawLine(this.lineStart, this.lineEnd, this.slits);
                pointCount = this.slits.Count - pointCount;

                this.pictureBox1.Refresh();
                this.updateTimeline(pointCount);
            }
            else if (this.slitTool == SlitTool.Squiggly) {
                this.drawSquiggly = false;
                this.bresenhamDrawLine(this.lineStart, e.Location, this.preSlit);

                int pointCount = this.preSlit.Count;
                this.slits.AddRange(this.preSlit);

                this.pictureBox1.Refresh();
                this.updateTimeline(pointCount);
            }
        }

        /// <summary>
        /// Updates the timeline by copying pixels from the source image into the timeline view.
        /// Note that we only update it for the newest pixel points, which were the last <i>pointsToUpdate</i>
        /// pixels in the slit tear pixel collection.
        /// </summary>
        /// <param name="pointsToUpdate">Number of pixels to actually update</param>
        protected void updateTimeline(int pointsToUpdate) {
            if (this.videoFrames == null) return;
            Bitmap bmp = (Bitmap) this.timelineForm.TimelineImage;
            
            for (int x = 0; x < this.videoFrames.Length; x++) {
                this.progressBar1.Value = x;
                BitmapManipulations.CopyFromPointListToColumn(this.videoFrames[x], bmp, x, pointsToUpdate, this.slits);
            }
            this.timelineForm.Refresh();
        }

        #region Line drawing Algorithm

        // Bresenham's line drawing algorithm, adapted from http://www.cs.unc.edu/~mcmillan/comp136/Lecture6/Lines.html
        // Unless its a rubber band line, add the point coordinates determined by the algorithm to our array of points
        // comprising all line segments seen so far
        public void bresenhamDrawLine(Point p0, Point p1, List<Point> points) {
            int x0 = p0.X;
            int y0 = p0.Y;
            int x1 = p1.X;
            int y1 = p1.Y;

            int dy = y1 - y0;
            int dx = x1 - x0;
            int stepx, stepy;

            if (dy < 0) { dy = -dy; stepy = -1; } else { stepy = 1; }
            if (dx < 0) { dx = -dx; stepx = -1; } else { stepx = 1; }
            dy <<= 1;                                                  // dy is now 2*dy
            dx <<= 1;                                                  // dx is now 2*dx

            points.Add(new Point(x0, y0));

            if (dx > dy) {
                int fraction = dy - (dx >> 1);                         // same as 2*dy - dx
                while (x0 != x1) {
                    if (fraction >= 0) {
                        y0 += stepy;
                        fraction -= dx;                                // same as fraction -= 2*dx
                    }
                    x0 += stepx;
                    fraction += dy;                                    // same as fraction -= 2*dy

                    points.Add(new Point(x0, y0));
                }
            }
            else {
                int fraction = dx - (dy >> 1);
                while (y0 != y1) {
                    if (fraction >= 0) {
                        x0 += stepx;
                        fraction -= dy;
                    }
                    y0 += stepy;
                    fraction += dx;

                    points.Add(new Point(x0, y0));
                }
            }
        }
        #endregion

        /// <summary>
        /// Resets the timeline
        /// </summary>
        private void button2_Click(object sender, EventArgs e) {
            this.resetTimeline();
        }

        /// <summary>
        /// Resets the timeline for real: clears the slit tear pixel collection, and
        /// tells the timeline form to fix itself up.
        /// </summary>
        protected void resetTimeline() {
            this.slits.Clear();
            this.pictureBox1.Refresh();
            this.timelineForm.TimelineImage = new Bitmap(this.timelineForm.TimelineImage.Width, this.timelineForm.TimelineImage.Height);
            this.timelineForm.Refresh();
        }

    }
}
